/** Temperature.java
 *  Class Invariant: myScale == 'C' && myDegrees >= ABSOLUTE_ZERO_CELSIUS
 *                || myScale == 'F' && myDegrees >= ABSOLUTE_ZERO_FAHRENHEIT
 *                || myScale == 'K'	&& myDegrees >= ABSOLUTE_ZERO_KELVIN 
 */
 
import ann.easyio.Keyboard;
import ann.util.Controller;

class Temperature
{
 public final static double ABSOLUTE_ZERO_FAHRENHEIT = -459.67,
                            ABSOLUTE_ZERO_CELSIUS    = -273.15,
                            ABSOLUTE_ZERO_KELVIN     =    0.0;

 //--- Constructors ---
 /** Default-value constructor.
  *  Postcondition: myDegrees == 0.0 && myScale == 'C'.
  */
 public Temperature()
 {
   myDegrees = 0.0;
   myScale = 'C';
 }

 /** Explicit-value constructor.
  *  Receive:       double degrees, char scale
  *  Precondition:  scale is one of 'f', 'F', 'c', 'C', 'k', or 'K' 
  *                  and degrees is a valid number of degrees for scale. 
  *  Postcondition: myDegrees == degrees && myScale == (uppercase) scale.
  */
 public Temperature(double degrees, char scale)
 {
   if (isValidTemperature(degrees, scale))
   {
     myDegrees = degrees;
     myScale = Character.toUpperCase(scale);
   }
   else
     fatal("Temperature(degrees, scale)", "invalid args: " + degrees + scale);
 }

 //--- Accessors ---
 /** degrees accessor.
  *  Return: myDegrees
  */
 public double getDegrees()
 {
   return myDegrees;
 }

 /** scale accessor.
  *  Return: myScale
  */
 public char getScale()
 {
   return myScale;
 } 
 
 //--- Input/Output ---
 /** read(Keyboard) using the ann Keyboard class.
  *  Receive:       theKeyboard, a Keyboard object
  *  Input:         degrees, a double, and scale, a char, from theKeyboard
  *  Return:        true if (degrees, scale) comprises a valid temperature, 
  *                  false otherwise
  *  Postcondition: myDegrees == degrees && myScale == scale if a valid
  *                  temperature has been input; otherwise the data members
  *                  are unchanged.
  */
  public boolean read(Keyboard theKeyboard)
  {
    double inDegrees = theKeyboard.readDouble();
    char inScale   = theKeyboard.readChar();

    if (isValidTemperature(inDegrees, inScale))
    {
      myDegrees = inDegrees;
      myScale   = Character.toUpperCase(inScale);
      return true;
    }
    else
      return false;
  }

 /** toString converter (used by +, print(), println(), ...).
  *  Return: a String representation of myself
  */
 public String toString()
 {
   return myDegrees + " " + myScale;
 }

//--- Converters ---
 /** Fahrenheit converter
  *  Return: the Fahrenheit equivalent to myself
  */
 public Temperature inFahrenheit()
 {
   Temperature result = null;

   if (myScale == 'F')
     result = new Temperature(myDegrees, 'F');
   else if (myScale == 'C')
     result = new Temperature(myDegrees * 1.8 + 32.0, 'F');
   else if (myScale == 'K')
     result = new Temperature((myDegrees - 273.15) * 1.8 + 32.0, 'F');
 
   return result;
 }

 /** Celsius converter stub
  */
 public Temperature inCelsius()
 {
   Temperature result = null;
   return result;
 }

/** Kelvin converter stub
 */
 public Temperature inKelvin()
 {
   Temperature result = null;
   return result;
 }

 //--- Other Temperature Operations
 /** raise by a (double) amount in degrees of my scale.
  *  Receive:      double amount
  *  Precondition: myDegrees + amount produces a valid magnitude
  *                 for a Temperature whose scale is myScale.
  *  Return:       a Temperature that is amount degrees higher than myself.
  */
 public Temperature raise(double amount)
 {
   double newDegrees = myDegrees + amount;

   if (!isValidTemperature(newDegrees, myScale))
     Controller.fatal("raise(double)", newDegrees + " " + myScale
                                  + " is not a valid temperature");

   return new Temperature(newDegrees, myScale);
 }

 /** compareTo utility method.
  *  Receive: otherTemp, the Temperature to which I'm being compared
  *  Return:  a negative value, if I am < otherTemp, 
  *           0,                if I am == otherTemp, 
  *           a positive value, if I am > otherTemp.
  */
 public double compareTo(Temperature otherTemp)
 {
   Temperature localTemp = null;          // the equivalent of otherTemp,
                                           //  but in my scale
   if (myScale == 'C')
     localTemp = otherTemp.inCelsius();
   else if (myScale == 'F')
     localTemp = otherTemp.inFahrenheit();
   else if (myScale == 'K')
     localTemp = otherTemp.inKelvin();
 
   if (myDegrees < localTemp.getDegrees()) 
     return -1.0;
   else if (myDegrees > localTemp.getDegrees())
     return 1.0;
   else
     return 0.0;
 }

 /** equality comparison.
  *  Receive: otherTemp, a Temperature object
  *  Return:  true if current Temperature object == otherTemp, false otherwise
  */
 public boolean equals(Temperature otherTemp)
 {
   return compareTo(otherTemp) == 0.0;
 }

/** less-than comparison.
  *  Receive: otherTemp, a Temperature object
  *  Return:  true if current Temperature object < otherTemp, false otherwise
  */ 
 public boolean lessThan(Temperature otherTemp)
 {
   return compareTo(otherTemp) < 0.0;
 }

 /** copy instance method, to support assignments, initializations.
  *  Return: A distinct copy of myself
  */
 public Temperature copy()
 {
   return new Temperature(myDegrees, myScale);
 }

 /** Temperature validation utility.
  * Receive: degrees, a double; scale, a char
  * Return:  true if (degrees, scale) represents a valid temperature,
  *           false otherwise.
  */
 public static boolean isValidTemperature(double degrees, char scale)
 {
   if (scale == 'C' || scale == 'c')
     return degrees >= ABSOLUTE_ZERO_CELSIUS;
   else if (scale == 'F' || scale == 'f')
     return degrees >= ABSOLUTE_ZERO_FAHRENHEIT;
   else if (scale == 'K' || scale == 'k')
     return degrees >= ABSOLUTE_ZERO_KELVIN;
   else
     return false;
 }

 /** Fatal error utility.
  *  Receive:       methodName, the name of the method where the error
  *                 occurred, and diagnostic, a message explaining the problem
  *  Output:        an error message containing methodName and diagnostic
  *  Postcondition: Execution has been terminated.
  */
 private static void fatal(String methodName, String diagnostic)
 {
   System.err.println("\n*** " + methodName + ": " + diagnostic);
   System.err.flush();
   System.exit(1); 
 }


//--- Attribute variables
 private double myDegrees;       // >= ABSOLUTE_ZERO for myScale
 private char   myScale;         // 'F', 'C', or 'K'
}

